import express, {
  Request,
  type Express,
  Response,
  NextFunction,
  Router,
} from "express";
import https from "node:https";
import { Readable } from "node:stream";
import fs from "node:fs";
import { MatrixError } from "./types/matrix";
import { MatrixHandler } from "./matrix";
import bodyParser from "body-parser";
import morgan from "morgan";
import { Discord } from "./discord";

class Webserver_ {
  app: Express;

  constructor() {
    this.app = express();

    this.app.use(morgan("tiny"));

    this.app.get("/", (req, res) => {
      res.send("matrix-discord-bridge " + (req.secure ? "using ssl" : ""));
    });

    this.setupMatrixRoutes();

    this.app.listen(process.env.PORT, () => {
      console.log("Listening on :" + process.env.PORT);
    });

    if (
      process.env.NODE_ENV === "development" &&
      !isNaN(parseInt(process.env.DEV_SSL_PORT!)) &&
      process.env.DEV_SSL_KEY_PATH &&
      process.env.DEV_SSL_CERT_PATH
    ) {
      if (
        !fs.existsSync(process.env.DEV_SSL_KEY_PATH) ||
        !fs.existsSync(process.env.DEV_SSL_CERT_PATH)
      ) {
        console.warn(
          "[DEV] SSL proxy could not start; couldn't find DEV_SSL_KEY_PATH or DEV_SSL_CERT_PATH"
        );
      } else {
        https
          .createServer(
            {
              key: fs.readFileSync(process.env.DEV_SSL_KEY_PATH),
              cert: fs.readFileSync(process.env.DEV_SSL_CERT_PATH),
            },
            this.app
          )
          .listen(
            parseInt(process.env.DEV_SSL_PORT!),
            "0.0.0.0",
            undefined,
            () => {
              console.log(
                "[DEV] SSL proxy running on :" + process.env.DEV_SSL_PORT
              );
            }
          );
      }
    }
  }

  checkAuth(req: Request, res: Response, next: NextFunction) {
    if (!req.headers.authorization)
      return res.status(401).json({ errcode: "M_MISSING_TOKEN" });

    const token = req.headers.authorization.replace("Bearer ", "");
    if (token !== process.env.MATRIX_HS_TOKEN)
      return res.status(403).json({ errcode: "M_UNAUTHORIZED" });

    next();
  }

  setupMatrixRoutes() {
    const router = Router();

    router.use(this.checkAuth);
    router.use(bodyParser.json());

    router.post("/ping", (req, res) => {
      res.status(200).json({});
    });

    router.put("/transactions/:txnId", async (req, res) => {
      try {
        const { events, "de.sorunome.msc2409.ephemeral": ephemeral } = req.body;

        await MatrixHandler.handleTransaction(
          req.params.txnId,
          events,
          ephemeral
        );

        res.status(200).json({});
      } catch (e) {
        // TODO: Error log
        console.error(req.url, e);
        res.status(500).json({});
      }
    });

    router.get("/users/:userId", async (req, res) => {
      try {
        const data = await MatrixHandler.handleUserQuery(req.params.userId);

        res.status(data.status).json("data" in data ? data.data : {});
      } catch (e) {
        // TODO: Error log
        console.error(req.url, e);
        res.status(500).json({});
      }
    });

    router.get("/rooms/:alias", async (req, res) => {
      try {
        const data = await MatrixHandler.handleRoomQuery(req.params.alias);

        res.status(data.status).json("data" in data ? data.data : {});
      } catch (e) {
        // TODO: Error log
        console.error(req.url, e);
        res.status(500).json({});
      }
    });

    // #region /thirdpart/{location,protocol} -- unused

    router.get("/thirdparty/location", async (req, res) => {
      // this shouldn't be called, but just in case return 404
      res.status(404).json({ errcode: "M_NOT_FOUND" });
    });

    router.get("/thirdparty/location/:protocol", async (req, res) => {
      // this shouldn't be called, but just in case return 404
      res.status(404).json({ errcode: "M_NOT_FOUND" });
    });

    router.get("/thirdparty/protocol/:protocol", async (req, res) => {
      // this shouldn't be called, but4 just in case return 404
      res.status(404).json({ errcode: "M_NOT_FOUND" });
    });

    router.get("/thirdparty/user", async (req, res) => {
      // this shouldn't be called, but4 just in case return 404
      return res.status(404).json({ errcode: "M_NOT_FOUND" });
    });

    router.get("/thirdparty/user/:protocol", async (req, res) => {
      // this shouldn't be called, but4 just in case return 404
      return res.status(404).json({ errcode: "M_NOT_FOUND" });
    });

    // #endregion

    router.use((req, res) => {
      res.status(404).json({
        errcode: "M_UNRECOGNIZED",
        error: "Unknown route",
      } satisfies MatrixError);
    });

    this.app.use("/_matrix/app/v1", router);

    this.app.get("/.well-known/matrix/server", (req, res) => {
      res.json({
        "m.server": process.env.HOST + ":443",
      });
    });

    this.app.get(
      "/_matrix/media/v3/download/:hostname/:name",
      async (req, res) => {
        if (req.params.hostname !== process.env.HOST) {
          return res.status(400).json({
            errcode: "M_ERROR",
            error: "This service does not proxy external MXCs",
          });
        }

        const parts = Buffer.from(req.params.name, "base64url")
          .toString()
          .split(",");

        const mode = parts.shift();
        let targetURL: string;

        switch (mode) {
          case "attachment": {
            const [guild_id, channel_id, message_id, attachment_id] = parts;

            const attachment = await Discord.getAttachmentURL(
              guild_id,
              channel_id,
              message_id,
              attachment_id
            );
            if (!attachment) {
              return res.status(404).json({});
            }

            targetURL = attachment;
            break;
          }
          case "avatar": {
            const [userid] = parts;

            const user = await Discord.client.users
              .fetch(userid)
              .catch((e) => null);
            if (!user) {
              return res.status(404).json({});
            }

            const avatarURL = user.avatarURL({
              extension: "png",
              forceStatic: true,
            });
            if (!avatarURL) {
              return res.status(404).json({});
            }

            targetURL = avatarURL;
            break;
          }
          case "sticker": {
            const [stickerid] = parts;

            const sticker = await Discord.client
              .fetchSticker(stickerid)
              .catch((e) => null);
            if (!sticker) {
              return res.status(404).json({});
            }

            targetURL = sticker.url;
            break;
          }
          case "emoji": {
            let [emojiid] = parts;
            let isAnimated = emojiid[0] === "a";
            if (isAnimated) emojiid = emojiid.slice(1);

            targetURL = `https://cdn.discordapp.com/emojis/${emojiid}.${isAnimated ? "gif" : "png"}`;
            break;
          }
          default:
            return res
              .status(404)
              .json({ errcode: "M_UNKNOWN", error: "Unknown media" });
        }

        if (req.query.allow_redirect) {
          // if this is set, the server/client accepts redirects
          return res.redirect(307, targetURL);
        }

        // redirects not supported, so we have to proxy it ourselves
        fetch(targetURL).then((proxy) => {
          if (proxy.status === 200 && proxy.body) {
            // this works; idk why ts gets mad
            Readable.fromWeb(proxy.body as any).pipe(res as any);
          } else {
            res.send("bad request");
          }
        });
      }
    );
  }
}

export const Webserver = new Webserver_();
